@@ -33,6 +33,7 @@ class PyTracer(object):
         self.data = None
         self.should_trace = None
         self.should_trace_cache = None
+        self.warn = None
         self.cur_file_data = None
         self.last_line = 0
         self.data_stack = []
@@ -43,8 +44,8 @@ class PyTracer(object):
     def _trace(self, frame, event, arg_unused):
         """The trace function passed to sys.settrace."""
 
-        #print "trace event: %s %r @%d" % (
-        #           event, frame.f_code.co_filename, frame.f_lineno)
+        #print("trace event: %s %r @%d" % (
+        #           event, frame.f_code.co_filename, frame.f_lineno))
 
         if self.last_exc_back:
             if frame == self.last_exc_back:
@@ -64,6 +65,8 @@ class PyTracer(object):
             if tracename is None:
                 tracename = self.should_trace(filename, frame)
                 self.should_trace_cache[filename] = tracename
+            #print("called, stack is %d deep, tracename is %r" % (
+            #               len(self.data_stack), tracename))
             if tracename:
                 if tracename not in self.data:
                     self.data[tracename] = {}
@@ -77,10 +80,10 @@ class PyTracer(object):
             # Record an executed line.
             if self.cur_file_data is not None:
                 if self.arcs:
-                    #print "lin", self.last_line, frame.f_lineno
+                    #print("lin", self.last_line, frame.f_lineno)
                     self.cur_file_data[(self.last_line, frame.f_lineno)] = None
                 else:
-                    #print "lin", frame.f_lineno
+                    #print("lin", frame.f_lineno)
                     self.cur_file_data[frame.f_lineno] = None
             self.last_line = frame.f_lineno
         elif event == 'return':
@@ -89,8 +92,9 @@ class PyTracer(object):
                 self.cur_file_data[(self.last_line, -first)] = None
             # Leaving this function, pop the filename stack.
             self.cur_file_data, self.last_line = self.data_stack.pop()
+            #print("returned, stack is %d deep" % (len(self.data_stack)))
         elif event == 'exception':
-            #print "exc", self.last_line, frame.f_lineno
+            #print("exc", self.last_line, frame.f_lineno)
             self.last_exc_back = frame.f_back
             self.last_exc_firstlineno = frame.f_code.co_firstlineno
         return self._trace
@@ -106,6 +110,10 @@ class PyTracer(object):
 
     def stop(self):
         """Stop this Tracer."""
+        if hasattr(sys, "gettrace") and self.warn:
+            if sys.gettrace() != self._trace:
+                msg = "Trace function changed, measurement is likely wrong: %r"
+                self.warn(msg % sys.gettrace())
         sys.settrace(None)
 
     def get_stats(self):
@@ -134,7 +142,7 @@ class Collector(object):
     # the top, and resumed when they become the top again.
     _collectors = []
 
-    def __init__(self, should_trace, timid, branch):
+    def __init__(self, should_trace, timid, branch, warn):
         """Create a collector.
 
         `should_trace` is a function, taking a filename, and returning a
@@ -150,8 +158,12 @@ class Collector(object):
         collecting data on which statements followed each other (arcs).  Use
         `get_arc_data` to get the arc data.
 
+        `warn` is a warning function, taking a single string message argument,
+        to be used if a warning needs to be issued.
+
         """
         self.should_trace = should_trace
+        self.warn = warn
         self.branch = branch
         self.reset()
 
@@ -191,6 +203,7 @@ class Collector(object):
         tracer.arcs = self.branch
         tracer.should_trace = self.should_trace
         tracer.should_trace_cache = self.should_trace_cache
+        tracer.warn = self.warn
         fn = tracer.start()
         self.tracers.append(tracer)
         return fn
