@@ -11,7 +11,7 @@ from coverage.data import CoverageData
 from coverage.files import FileLocator, TreeMatcher, FnmatchMatcher
 from coverage.files import find_python_files
 from coverage.html import HtmlReporter
-from coverage.misc import CoverageException, bool_or_none
+from coverage.misc import CoverageException, bool_or_none, join_regex
 from coverage.results import Analysis, Numbers
 from coverage.summary import SummaryReporter
 from coverage.xmlreport import XmlReporter
@@ -64,11 +64,15 @@ class coverage(object):
         measured.
 
         `include` and `omit` are lists of filename patterns. Files that match
-        `include` will be measured, files that match `omit` will not.
+        `include` will be measured, files that match `omit` will not.  Each
+        will also accept a single string argument.
 
         """
         from coverage import __version__
 
+        # A record of all the warnings that have been issued.
+        self._warnings = []
+
         # Build our configuration from a number of sources:
         # 1: defaults:
         self.config = CoverageConfig()
@@ -92,6 +96,10 @@ class coverage(object):
             self.config.data_file = env_data_file
 
         # 4: from constructor arguments:
+        if isinstance(omit, string_class):
+            omit = [omit]
+        if isinstance(include, string_class):
+            include = [include]
         self.config.from_args(
             data_file=data_file, cover_pylib=cover_pylib, timid=timid,
             branch=branch, parallel=bool_or_none(data_suffix),
@@ -101,8 +109,10 @@ class coverage(object):
         self.auto_data = auto_data
         self.atexit_registered = False
 
-        self.exclude_re = ""
-        self._compile_exclude()
+        # _exclude_re is a dict mapping exclusion list names to compiled
+        # regexes.
+        self._exclude_re = {}
+        self._exclude_regex_stale()
 
         self.file_locator = FileLocator()
 
@@ -115,12 +125,12 @@ class coverage(object):
             else:
                 self.source_pkgs.append(src)
 
-        self.omit = self._abs_files(self.config.omit)
-        self.include = self._abs_files(self.config.include)
+        self.omit = self._prep_patterns(self.config.omit)
+        self.include = self._prep_patterns(self.config.include)
 
         self.collector = Collector(
             self._should_trace, timid=self.config.timid,
-            branch=self.config.branch
+            branch=self.config.branch, warn=self._warn
             )
 
         # Suffixes are a bit tricky.  We want to use the data suffix only when
@@ -223,14 +233,19 @@ class coverage(object):
         self._check_for_packages()
 
         # Compiled Python files have two filenames: frame.f_code.co_filename is
-        # the filename at the time the .pyc was compiled.  The second name
-        # is __file__, which is where the .pyc was actually loaded from.  Since
+        # the filename at the time the .pyc was compiled.  The second name is
+        # __file__, which is where the .pyc was actually loaded from.  Since
         # .pyc files can be moved after compilation (for example, by being
         # installed), we look for __file__ in the frame and prefer it to the
         # co_filename value.
         dunder_file = frame.f_globals.get('__file__')
         if dunder_file:
             filename = self._source_for_file(dunder_file)
+
+        # Jython reports the .class file to the tracer, use the source file.
+        if filename.endswith("$py.class"):
+            filename = filename[:-9] + ".py"
+
         canonical = self.file_locator.canonical_filename(filename)
 
         # If the user specified source, then that's authoritative about what to
@@ -261,7 +276,7 @@ class coverage(object):
     # To log what should_trace returns, change this to "if 1:"
     if 0:
         _real_should_trace = _should_trace
-        def _should_trace(self, filename, frame):   # pylint: disable-msg=E0102
+        def _should_trace(self, filename, frame):   # pylint: disable=E0102
             """A logging decorator around the real _should_trace function."""
             ret = self._real_should_trace(filename, frame)
             print("should_trace: %r -> %r" % (filename, ret))
@@ -269,12 +284,27 @@ class coverage(object):
 
     def _warn(self, msg):
         """Use `msg` as a warning."""
-        sys.stderr.write("Coverage.py warning: " + msg + "\n")
+        self._warnings.append(msg)
+        sys.stderr.write("Coverage.py warning: %s\n" % msg)
 
-    def _abs_files(self, files):
-        """Return a list of absolute file names for the names in `files`."""
-        files = files or []
-        return [self.file_locator.abs_file(f) for f in files]
+    def _prep_patterns(self, patterns):
+        """Prepare the file patterns for use in a `FnmatchMatcher`.
+
+        If a pattern starts with a wildcard, it is used as a pattern
+        as-is.  If it does not start with a wildcard, then it is made
+        absolute with the current directory.
+
+        If `patterns` is None, an empty list is returned.
+
+        """
+        patterns = patterns or []
+        prepped = []
+        for p in patterns or []:
+            if p.startswith("*") or p.startswith("?"):
+                prepped.append(p)
+            else:
+                prepped.append(self.file_locator.abs_file(p))
+        return prepped
 
     def _check_for_packages(self):
         """Update the source_match matcher with latest imported packages."""
@@ -295,7 +325,7 @@ class coverage(object):
                 try:
                     pkg_file = mod.__file__
                 except AttributeError:
-                    self._warn("Module %s has no python source." % pkg)
+                    self._warn("Module %s has no Python source." % pkg)
                 else:
                     d, f = os.path.split(pkg_file)
                     if f.startswith('__init__.'):
@@ -367,30 +397,49 @@ class coverage(object):
         self.collector.reset()
         self.data.erase()
 
-    def clear_exclude(self):
+    def clear_exclude(self, which='exclude'):
         """Clear the exclude list."""
-        self.config.exclude_list = []
-        self.exclude_re = ""
+        setattr(self.config, which + "_list", [])
+        self._exclude_regex_stale()
 
-    def exclude(self, regex):
+    def exclude(self, regex, which='exclude'):
         """Exclude source lines from execution consideration.
 
-        `regex` is a regular expression.  Lines matching this expression are
-        not considered executable when reporting code coverage.  A list of
-        regexes is maintained; this function adds a new regex to the list.
-        Matching any of the regexes excludes a source line.
+        A number of lists of regular expressions are maintained.  Each list
+        selects lines that are treated differently during reporting.
+
+        `which` determines which list is modified.  The "exclude" list selects
+        lines that are not considered executable at all.  The "partial" list
+        indicates lines with branches that are not taken.
+
+        `regex` is a regular expression.  The regex is added to the specified
+        list.  If any of the regexes in the list is found in a line, the line
+        is marked for special treatment during reporting.
 
         """
-        self.config.exclude_list.append(regex)
-        self._compile_exclude()
+        excl_list = getattr(self.config, which + "_list")
+        excl_list.append(regex)
+        self._exclude_regex_stale()
 
-    def _compile_exclude(self):
-        """Build the internal usable form of the exclude list."""
-        self.exclude_re = "(" + ")|(".join(self.config.exclude_list) + ")"
+    def _exclude_regex_stale(self):
+        """Drop all the compiled exclusion regexes, a list was modified."""
+        self._exclude_re.clear()
 
-    def get_exclude_list(self):
-        """Return the list of excluded regex patterns."""
-        return self.config.exclude_list
+    def _exclude_regex(self, which):
+        """Return a compiled regex for the given exclusion list."""
+        if which not in self._exclude_re:
+            excl_list = getattr(self.config, which + "_list")
+            self._exclude_re[which] = join_regex(excl_list)
+        return self._exclude_re[which]
+
+    def get_exclude_list(self, which='exclude'):
+        """Return a list of excluded regex patterns.
+
+        `which` indicates which list is desired.  See `exclude` for the lists
+        that are available, and their meaning.
+
+        """
+        return getattr(self.config, which + "_list")
 
     def save(self):
         """Save the collected coverage data to the data file."""
@@ -432,7 +481,7 @@ class coverage(object):
             # If there are still entries in the source_pkgs list, then we never
             # encountered those packages.
             for pkg in self.source_pkgs:
-                self._warn("Source module %s was never encountered." % pkg)
+                self._warn("Module %s was never imported." % pkg)
 
             # Find out if we got any data.
             summary = self.data.summary()
@@ -487,7 +536,7 @@ class coverage(object):
         return Analysis(self, it)
 
     def report(self, morfs=None, show_missing=True, ignore_errors=None,
-                file=None,                          # pylint: disable-msg=W0622
+                file=None,                          # pylint: disable=W0622
                 omit=None, include=None
                 ):
         """Write a summary report to `file`.
