@@ -1,5 +1,6 @@
 """File wrangling."""
 
+from coverage.backward import to_string
 import fnmatch, os, sys
 
 class FileLocator(object):
@@ -72,9 +73,7 @@ class FileLocator(object):
                     data = zi.get_data(parts[1])
                 except IOError:
                     continue
-                if sys.version_info >= (3, 0):
-                    data = data.decode('utf8') # TODO: How to do this properly?
-                return data
+                return to_string(data)
         return None
 
 
